;;;  -*- mode: LISP; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Dan Bothell
;;; Copyright   : (c) 2009 Dan Bothell
;;; Availability: Covered by the GNU LGPL, see LGPL.txt
;;; Address     : Department of Psychology 
;;;             : Carnegie Mellon University
;;;             : Pittsburgh, PA 15213-3890
;;;             : db30@andrew.cmu.edu
;;; 
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : sf-module.lisp
;;; Version     : 1.1
;;; 
;;; Description : Module and device for connecting to the hacked synchronous python SF
;;;               game via a socket connection along with the code to run the model
;;;               and game in that way.
;;;
;;; Bugs        : 
;;;
;;; To do       : 
;;; 
;;; ----- History -----
;;;
;;; 2009.03.26 Dan
;;;             : * Initial version with "nothing" in it...
;;; 2009.12.01 Dan [1.1]
;;;             : * Stable version that can only handle the orbit condition
;;;             :   for now since all it sees is the ship itself.
;;; 2009.12.04 Dan 
;;;             : * Changing the slack hook to just schedule update-world
;;;             :   everytime since that's safe to do now and that will cause
;;;             :   conflict-resolution to "move back" when necessary.
;;; 2010.01.25 Dan
;;;             : * Added the code to load the other modules used at the end
;;;             :   instead of assuming that they're in place when ACT-R itself
;;;             :   is loaded -- which means the "speculative" modules need to
;;;             :   be in the right place.
;;; 2010.01.28 Dan
;;;             : * Added a wait to the connection code to make sure we get the
;;;             :   initial time before starting and also adding a game update
;;;             :   right at the start of the run now.
;;;             : * In addition, changed the time at which the slack schedules
;;;             :   the update because otherwise we're behind a tick in perception
;;;             :   because of how the game handles things (our 'ready' signal
;;;             :   is an action it processes before sending us the data which
;;;             :   will be followed by the next stopping time).
;;; 2010.02.04 Dan
;;;             : * Added the fortress, explosion, bonus, and vlner objects to
;;;             :   be seen by the model and changed it so that some items are
;;;             :   now "dynamic" meaning they may or may not be visible in 
;;;             :   every update.  The assumption is that an item is not visible
;;;             :   if it is dynamic unless the update has an instance for it.
;;; 2010.02.11 Dan
;;;             : * When the fortress explodes make that visible to the model --
;;;             :   assume if there was a fortress but isn't now then a 
;;;             :   fortress-explosion is to be shown.
;;;             :   
;;; 2010.02.15 Dan
;;;             : * Tweak the kind of the bonus item so that the visual-search
;;;             :   buffer can detect when it's new since the finsts (attended)
;;;             :   info isn't available in the chunks to test.
;;; 2010.02.22 Dan 
;;;             : * Make the update-world event non-maintenance because if the
;;;             :   model isn't actively tracking (moving attention to the bonus
;;;             :   item for example) then there are no model events to bring
;;;             :   conflict-resolution "back" to and it can still get stuck
;;;             :   with a long delay from temporal or declarative.
;;; 2010.02.23 Dan
;;;             : * Added the 2-handed and attend-and-track loaders to the end.
;;; 2010.03.01 Dan
;;;             : * Changed get-internal-real-time to (win::gettickcount) in
;;;             :   the logging of the socket data for the higher resolution
;;;             :   that it provides.  This does make it ACL specific for now.
;;; 2010.03.04 Dan
;;;             : * Modified show-ship-trial to display the mine as well when
;;;             :   it exists.
;;; 2010.03.11 Dan
;;;             : * Fixed some issues with a fortress explosion being "visible"
;;;             :   at the start of a trial if it was available on the previous
;;;             :   one and the ship's explosion continuing to be seen during
;;;             :   the 1 second of down-time.
;;; 2010.03.15 Dan
;;;             : * Adding a slot to the ship to indicate when it hits the 
;;;             :   fortress hex.  That will stay set until there is a change
;;;             :   to the ships velocity vector.
;;; 2010.03.19 Dan
;;;             : * Fixed the fortress location info -- it's at 355,315.
;;;             : * Added code to check whether or not a ship's shot will hit
;;;             :   the mine and to do the prediction to see if it will ever
;;;             :   be in a position to shoot it given it's current drift and
;;;             :   orientation or if it will need to turn.
;;; 2010.03.26 Dan
;;;             : * Scheduling the proc-display calls in the vis-loc-to-obj 
;;;             :   method because otherwise bad things occur with respect to
;;;             :   chunk deletion...
;;; 2010.03.31 Dan
;;;             : * Modifying the mine side to be ahead/behind instead of
;;;             :   left/right/ahead so the model can be more sensitive
;;;             :   to ones coming from behind.  The prediction code now also
;;;             :   computes a turn to both sides and stops at the closer one.
;;;             : * Now also determines when it needs to start dealing with
;;;             :   the mine based using the distance as the key since the
;;;             :   temporal is already in use.  The prediction has a slot for
;;;             :   the mine distance at which the model needs to switch focus
;;;             :   to the mine based on the "soonest" it can shoot it if the
;;;             :   ship doesn't change velocity.
;;; 2010.04.01 Dan
;;;             : * Decreasing the preference to drift from the mine prediction
;;;             :   code.  Now the drift only is if it's 9 or fewer ticks and
;;;             :   the drift instead is if it's a dif of 5.
;;; 2010.04.27 Dan
;;;             : * Fixed an issue with shooting at the mine where the prediction
;;;             :   could have it turn the wrong way when it comes in at about 90
;;;             :   degrees to the ship orientation.
;;; 2010.05.05 Dan
;;;             : * Modified update-world so that the fortress-hit is only
;;;             :   visible when there is still a fortress around.
;;; 2010.05.13 Dan
;;;             : * To improve running performance added the redefinition of
;;;             :   the ACT-R model structure to tweak the hash-table settings.
;;; 2010.06.15 Dan
;;;             : * Adjusted the mine prediction code to increase the action
;;;             :   time for imaginal to be 300ms so, now assume 12 ticks before
;;;             :   any turning.
;;; 2010.06.17 Dan
;;;             : * Adjusted the mine prediction code to increase the action
;;;             :   time for imaginal to be 250ms so, now assume 10 ticks before
;;;             :   any turning.
;;; 2010.06.23 Dan
;;;             : * Reverted to a 200ms action for the prediction and added 
;;;             :   a new imaginal action for returning to orbit after shooting
;;;             :   a mine.
;;; 2010.09.10 Dan
;;;             : * Added a speed-change slot to the ship and changed the 
;;;             :   "reassess" distance (the hit-dist) from hit -25 to hit - 50
;;;             :   so the model waits a little longer before trying to fix
;;;             :   things.
;;; 2010.09.17 Dan
;;;             : * Changed the reassess from -50 to -90 to try waiting even
;;;             :   a little more...
;;; 2010.11.24 Dan
;;;             : * Cleaned up a few things for distribution.  Basically, removed
;;;             :   the ACL specific code for recording events and drawing debuging
;;;             :   info eventhough nobody would have been using that anyway.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

#+:packaged-actr (in-package :act-r)
#+(and :clean-actr (not :packaged-actr) :ALLEGRO-IDE) (in-package :cg-user)
#-(or (not :clean-actr) :packaged-actr :ALLEGRO-IDE) (in-package :cl-user)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Redefine the ACT-R model structure so that the hash-table is more appropriate
;;; for this task -- reduces the run time for a day from ~6 minutes to just over
;;; 2 with (play-sf (* 210.01 16) :speed 150).

(defstruct act-r-model
  "The internal structure of a model"
  (modules-table (make-hash-table)) 
  (buffers (make-hash-table))
  (chunks-table (make-hash-table :test 'eq  :rehash-threshold .6 :size 18000))
  (chunk-ref-table (make-hash-table :test 'eq  :rehash-threshold .6 :size 18000))
  (chunk-types-table (make-hash-table :test 'eq))
  (largest-chunk-type 0)
  name 
  code
  (chunk-update t)
  (dynamic-update t)
  delete-chunks
  dynamic-update-hooks
  short-copy-names
  ;device
  )


(require-compiled "UNI-FILES" "ACT-R6:support;uni-files")

(defparameter *host-address* "127.0.0.1" "Address/name of the machine running the server")
(defparameter *port-for-server* 3000 "default port for the server") 

;;;;;;;;;;;;;;;;;;;;;
;;; Tools 

;;; Communication to the game is via a 5 character long string terminated with
;;; a carriage return linefeed pair.  The sending functions are responsible for
;;; ensuring the 5 character size at this point.

(defconstant *cr/lf* (format nil "~C~C" #\cr #\lf))

(defmacro socket-write-line (world string &rest args)
  "Write a CRLF-terminated line"
  `(if (or (null (socket ,world))
           (uni-stream-closed (socket ,world)))
       (print-warning "Can't send command {~S} because connection closed" 
                   (format nil ,string ,@args))
     (let ((string-to-write (format nil ,string ,@args)))
       (when (record-outgoing ,world)
         (push (cons (get-internal-real-time) string-to-write)
               (outgoing-record ,world)))
         
       (uni-send-string (socket ,world) 
                        (concatenate 'string string-to-write *cr/lf*)))))


;; This prints out the input and output communications together with a zeroed time
;; reference from the first message.
;; Useful for debugging connection/synchronization issues.

(defun interleave-logs (&optional (speed 1.0))
  (let* ((sf (get-module :sf))
         (log-list (stable-sort (append (outgoing-record sf) (reverse (incoming-record sf))) #'< :key #'car))
         (start-time (caar log-list))
         (zeroed-times (mapcar (lambda (x) (cons (/ (- (car x) start-time) (/ 1000.0 speed)) (cdr x))) log-list)))
    (dolist (x zeroed-times)
      (format t "~9,3f  ~s~%" (car x) (cdr x)))
    zeroed-times))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The UT device code

;;; Define a class for the device which will also be the instance for the module.

(defclass sf-world () 
  ((host :initform *host-address* :accessor sf-host)
   (port :initform *port-for-server* :accessor sf-port)
   
   (state-process :initform nil :accessor state-process)
   (socket :initform nil :accessor socket)
   (game-state :initform nil :accessor game-state)
   
   (last-proc-handled :initform t :accessor last-proc-handled)
   
   (collect-incoming :initform nil :accessor collect-incoming)
   (incoming-record :initform nil :accessor incoming-record)
   (record-outgoing :initform nil :accessor record-outgoing)
   (outgoing-record :initform nil :accessor outgoing-record)
   
   
   (show-visual :initform nil :accessor show-visual)
   
   ;; Other slots as needed...
   
   (all-locs :initform nil :accessor all-locs)
   (dynamics :initform nil :accessor dynamics)
   (loc->obj-table :initform (make-hash-table :test 'eq) :accessor loc->obj-table)
   
   (end-time :initform 0 :accessor end-time)
   (last-time :initform 0 :accessor last-time)
   (start-time :initform nil :accessor start-time)
   (last-stop-time :initform nil :accessor last-stop-time)
   (sim-speed :initform 1 :accessor sim-speed)
   (proc-now :initform nil :accessor proc-now)
   (new-data :initform nil :accessor new-data)
   (fortress-there :initform nil :accessor fortress-there)
   (f-hit-counter :initform nil :accessor f-hit-counter)
   (bonus-timer :initform nil :accessor bonus-timer)
   (hit-hex :initform nil :accessor hit-hex)
   (last-v :initform nil :accessor last-v)
   (mine-visible :initform nil :accessor mine-visible)
   (fort-exists :initform nil :accessor fort-exists)
   ))
  

;;; Methods which need to be defined for the device.
;;; All methods should be provided even if not needed
;;; (the cursor and speech methods).

(defmethod device-move-cursor-to ((world sf-world) loc))
(defmethod device-handle-click ((world sf-world)))
(defmethod get-mouse-coordinates ((world sf-world)) (vector 0 0))
(defmethod device-speak-string ((world sf-world) text))

;;; The normal keypress method isn't being used since it
;;; doesn't have the necessary "up" notice after striking a key
;;; since the game responds to held keys.

(defmethod device-handle-keypress ((world sf-world) key)
  (model-output "Standard keypressing requests aren't handled.  Must use hold/release requests to have a duration.")
  )

;;; Instead the keypress module's hold and release methods are defined.
;;; To keep things simple it's just assuming the hand positions are over
;;; the correct keys.  The model may set them that way if it wants, but
;;; it will not affect how it plays.

(defparameter *suppress-thrust* nil)

(defun hand-finger-to-sf-key (hand finger)
  (case hand
    (left
     (case finger
       (middle (if *suppress-thrust* #\z #\w))
       (ring   #\a)
       (index  #\d)
       (thumb  #\space)))
    (right
     (case finger
       (index  #\j)
       (middle #\k)
       (ring   #\l)
       (thumb  #\space)))))


(defmethod device-hold-finger ((world sf-world) hand finger) 
  (socket-write-line world (format nil "~c2   " (hand-finger-to-sf-key hand finger))))

(defmethod device-release-finger ((world sf-world) hand finger)
  (socket-write-line world (format nil "~c3   " (hand-finger-to-sf-key hand finger))))


;;; The model's visual world is just maintained as a list of the chunks which are
;;; modified in the message handler.  So, this code is mostly trivial.  The only
;;; thing to worry about is signaling that it has processed things so that pending
;;; messages are treated appropriately (old ones can be pitched if new ones come
;;; in before the old ones were processed).

(defmethod build-vis-locs-for ((world sf-world) (vis vision-module))
  (uni-without-interrupts
   (setf (last-proc-handled world) t)
   ; (all-locs world)
   ;; This is really what will be needed once we allow for dynamic things
   ;; like mines, ship destruction, bonus letter, etc.
   (mapcan (lambda (x) (if (item-visible x) (list x) nil)) (all-locs world)))
   )

(defmethod item-visible (loc)
  (fast-chunk-slot-value-fct loc 'visible))


;;; Keep the visual objects in a simple look-up table -- since all the items in
;;; the world are known in advance they can be fixed chunks created at module start
;;; time.

(defmethod vis-loc-to-obj ((world sf-world) vis-loc)
  (when (eq vis-loc 'bonus-loc)
    (mod-chunk-fct 'bonus-loc '(kind bonus))
    (schedule-event-relative 0 'proc-display :maintenance t :module :sf))
  (when (eq vis-loc 'trial-end-loc)
    (mod-chunk-fct 'trial-end-loc '(kind trial-end))
    (schedule-event-relative 0 'proc-display :maintenance t :module :sf))
  
  (when (and (eq vis-loc 'explosion-loc) (eq (chunk-slot-value explosion-loc kind) 'explosion))
    (mod-chunk-fct 'explosion-loc '(kind debris))
    (schedule-event-relative 0 'proc-display :maintenance t :module :sf))
  
  (gethash vis-loc (loc->obj-table world)))
  
(defvar *hex-points* '((155 . 315) (255 . 90) (455 . 90) (555 . 315) (455 . 540) (255 . 540)))



;;; These functions were for trying out some alternative distance measures in an attempt
;;; to get the model to fly the oval subjects seem to in the orbit conditions, but for
;;; now is not being used.

#| This computes the shortest perpendicular distance from a point to one of the hex sides.
|#

(defun closest-side (x y)
  (apply 'min (mapcar (lambda (p1 p2)
                        (point-to-line x y (car p1) (cdr p1) (car p2) (cdr p2)))
                *hex-points* (append (cdr *hex-points*) (list (car *hex-points*))))))

(defun point-to-line (m n x1 y1 x2 y2)
              (let* ((b 1)
                     (s (/ (- y2 y1) (- x2 x1)))
                     (a (- s))
                     (c (- (* x1 s) y1)))
                (/ (abs (+ (* a m) (* b n) c)) (sqrt (+ (* a a) (* b b))))))


;; just hard code the checks instead of doing something fancy
;; take the line segments in order using hex-points like in closest-side
;; which gives the equations of ax + by + c = 0 of:
#|
 (2.25 1.0 -663.75)
 (0.0 1.0 -90.0)
 (-2.25 1.0 933.75)
 (2.25 1.0 -1563.75)
 (0.0 1.0 -540.0)
 (-2.25 1.0 33.75)
|#

;;; This computes the distance to hit the nearest side along the ships movement
;;; vector.

(defun hex-hit-dist (x y vx vy) 
     (cond ((and (zerop vx) (zerop vy))
           (closest-side x y))
          (t
           (let* ((vy (- vy))  ;; comes over with up being positive even though the screen
                               ;; coords are the other way
                  (dir (rad->deg (atan vy vx))) ;; and degrees are easier for me to specify below
                 
                 (min 1000))
             
            ; (format t "dir = ~F~%" dir)
             
             (when (or (> dir 120) (< dir -60))
               (let ((d (intersect-point-slope-line x y vx vy 2.25 1.0 -663.75)))
                 (when (< d min)
                   (setf min d))))
             
             (when (< dir 0)
               (let ((d (intersect-point-slope-line x y vx vy 0 1.0 -90)))
                 (when (< d min)
                   (setf min d))))
             
             (when (< -120 dir 60)
               (let ((d (intersect-point-slope-line x y vx vy -2.25 1.0 933.75)))
                 (when (< d min)
                   (setf min d))))
             
             (when (< -60 dir 120)
               (let ((d (intersect-point-slope-line x y vx vy 2.25 1.0 -1563.75)))
                 (when (< d min)
                   (setf min d))))
             
             (when (> dir 0)
               (let ((d (intersect-point-slope-line x y vx vy 0 1.0 -540.0)))
                 (when (< d min)
                   (setf min d))))
             
             (when (or (> dir 60) (< dir -120))
               (let ((d (intersect-point-slope-line x y vx vy -2.25 1.0 33.75)))
                 (when (< d min)
                   (setf min d))))
             min))))

(defun intersect-point-slope-line (x1 y1 vx vy a b c)
  (if (< vx .001)  ;; avoid issues with small but non-zero numbers
      (let* ((x x1)
             (y (/ (- (- (* a x)) c) b)))
        (dist (list x1 y1) (list x y)))
    (let* ((m (/ (- vy) vx))
           (x (/ (- (* b m x1) (* b y1) c) (+ a (* b m))))
           (y (+ (* m (- x x1)) y1)))
      (dist (list x1 y1) (list x y)))))
  

(defvar *show-ship-update* nil)

;;; This function is responsible for updating all of the visual and other
;;; game state information while the model is running.  This gets called
;;; from the slack hook whenever there's a need to update things -- it
;;; may not be called for all incoming data because if the model is going
;;; to skip past an update that info can be purged and not dealt with.
;;; !! Need to get trickier with that when there's dynamic info because
;;; the model may need the notices of things changing even when the update
;;; won't "matter".
;;; Right now all it does is update the ship chunk.


(defun update-world (world)
  
  (let ((bonus-on (fast-chunk-slot-value-fct 'bonus-loc 'visible)))
  
    ;; mark everything as not visible
    (mapc (lambda (x) (set-chunk-slot-value-fct x 'visible nil)) (dynamics world))
    
    ;; mark the mine as non-hitable and very far away because it will be recalculated
    (set-chunk-slot-value my-ship shoot-at-mine nil)
    (set-chunk-slot-value my-ship hit-fortress-first nil)
    (set-chunk-slot-value my-ship mine-dist 10000)
    
        
  ;; make sure to mark things visible
  
  (dolist (data (new-data world))
    (when (show-visual world) (pprint data))
    (case (car data)
      (:ship 
       (destructuring-bind (dummy x y angle vel dist vdir orient vx vy) data
         
         (cond ((null (last-v world))
                (setf (last-v world) (cons vx vy)))
               ((and (= vx (car (last-v world))) (= vy (cdr (last-v world))))
                ;; do nothing
                )
               (t
                (if (hit-hex world)
                    (setf (hit-hex world) nil)
                  (let* ((od (atan (cdr (last-v world)) (car (last-v world))))
                         (cd (atan vy vx))
                         (dif (abs (- cd od))))
                    (when (and (> dif 3) (< dif 3.3) (< dist 60)) ;; close to 180 shift and near hex
                      
                      ;(format t "~%OD: ~F ~F ~F  CD:  ~F ~F ~F~%~%"
                      ;  od (car (last-v world)) (cdr (last-v world)) cd vy vx)
                      
                      (setf (hit-hex world) t)))
                      
                    )
                (setf (car (last-v world)) vx)
                (setf (cdr (last-v world)) vy)))
         
         (when *show-ship-update*
           (format t "x: ~4d y: ~4d angle: ~3d vel: ~5,3f vx: ~5,3f vy: ~5,3f dist: ~3d vdir: ~3d~%" x y angle vel vx vy dist vdir))
         
         (let ((last-vel (chunk-slot-value my-ship vel)))
           (when (null last-vel)
             (setf last-vel 0))
           (mod-chunk-fct 'my-ship (list 'x x
                                       'y y
                                       'vx vx
                                       'vy vy
                                       'angle angle
                                       'vel vel
                                       'dist dist
                                       'vdir (if (= vel 0.0) angle vdir)
                                       'orientation orient
                                       ;; not using this right now
                                       'hex-dist   (hex-hit-dist x y vx vy)  ;(closest-side x y)
                                       ;; for motion purposes record last distance
                                       ;; gives a crude check for heading
                                       'last-dist (if (numberp (chunk-slot-value-fct 'my-ship 'dist))
                                                      (chunk-slot-value-fct 'my-ship 'dist)
                                                    dist)
                                       'speed-change (cond ((= vel last-vel) 0) ((< vel last-vel) -1) (t 1))
                                       'hex-hit (hit-hex world)
                                       )))
         (mod-chunk-fct 'own-ship-loc (list 'screen-x x 'screen-y y 'visible t))))
      
      (:explosion 
       (destructuring-bind (dummy x y) data
         (mod-chunk-fct 'explosion-loc (list 'screen-x x 'screen-y y 'visible t 'kind 'explosion)))
       (setf (hit-hex world) nil)
       (setf (last-v world) nil)
       )
      
      (:mine 
       (mod-chunk-fct 'mine-loc (list 'visible t 'kind 'mine))
       (let* ((x (second data))
              (y (third data))
              (o (chunk-slot-value my-ship orientation))
              (sx (chunk-slot-value my-ship x))
              (sy (chunk-slot-value my-ship y))
              (ma (rad->deg (atan (- sy y) (- x sx))))
              (a (- o ma))
              (side 'ahead)
              (dist (dist (list x y) (list sx sy)))
              (hit-dist (abs (* (sin (deg->rad a)) dist)))
              (fa (chunk-slot-value my-ship angle))
              (fd (chunk-slot-value my-ship dist))
              
              (hit-f-dist (abs (* (sin (deg->rad fa)) fd))))
         
         
         ;(format t "mx: ~s my: ~s  sx: ~s sy: ~s  o: ~s  ma: ~s  a: ~s  dist:  ~s  hit-dist: ~s~%"
         ;  x y sx sy o ma a dist hit-dist)
         
         (when (> a 180)
           (setf a (- a 360)))
         
         (when (< a -180)
           (setf a (+ a 360)))
         
         (unless (<= -90 a 90)
           (setf side 'behind))
         
         (mod-chunk-fct 'my-ship (list 'mine-side side))
               
         (mod-chunk-fct 'mine (list 'x x 'y y))
         (mod-chunk-fct 'mine-loc (list 'screen-x x 'screen-y y))
         
         (mod-chunk-fct 'my-ship (list 'mine-dist dist 'shoot-at-mine (<= hit-dist 20) 'hit-fortress-first (and (<= hit-f-dist 18) (< fd dist))))
         
         (setf (mine-visible world) 12)))

      
      (:fortress 
       (when (fort-exists world) ;; corrects for a bug that I don't want to
                                 ;; track down on the python side
                                 ;; If the ship is destroyed the fortress
                                 ;; indicator gets sent even in the no fortress
                                 ;; conditions (mines only since that's the only
                                 ;; one where the ship can be destroyed)
         (mod-chunk-fct 'fortress-loc (list 'visible t))
         (setf (fortress-there world) t)))
      (:fortress-hit
       
       (setf (f-hit-counter world) 0))
      (:bonus 
       (mod-chunk-fct 'bonus-loc (list 'visible t))
       (if bonus-on
           (when (> (- (mp-time) (bonus-timer world)) .5)
             (mod-chunk-fct 'bonus-loc '(kind bonus)))
         (progn
           (setf (bonus-timer world) (mp-time))
           (mod-chunk-fct 'bonus-loc '(kind new-bonus))))
       (mod-chunk-fct 'bonus (list 'value (second data))))
      (:vlner 
       (mod-chunk-fct 'vlner-loc (list 'visible t))
       (mod-chunk-fct 'vlner (list 'value (second data)))
       (setf (chunk-real-visual-value 'vlner-loc) (second data)))
      
      (:iff 
       (mod-chunk-fct 'iff-loc (list 'visible t))
       (mod-chunk-fct 'iff (list 'value (second data)))
       (setf (chunk-real-visual-value 'iff-loc) (second data)))
      
      
      (:trial-end
       (setf (fortress-there world) nil)
       (mod-chunk-fct 'trial-end-loc (list 'visible t 'kind 'new-trial-end))
       (setf (hit-hex world) nil)
       (setf (last-v world) nil)
       )
      (:foes
       
       ;; hack to make sure fortress isn't considered blown up
       ;;
       
       (setf (fortress-there world) nil)
       
       (mod-chunk-fct 'foe1-loc (list 'visible t))
       (setf (chunk-real-visual-value 'foe1-loc) (string (second data)))
       (mod-chunk-fct 'foe1 (list 'value (string (second data))))
       (mod-chunk-fct 'foe2-loc (list 'visible t))
       (setf (chunk-real-visual-value 'foe2-loc) (string (third data)))
       (mod-chunk-fct 'foe2 (list 'value (string (third data))))
       (mod-chunk-fct 'foe3-loc (list 'visible t))
       (setf (chunk-real-visual-value 'foe3-loc) (string (fourth data)))
       (mod-chunk-fct 'foe3 (list 'value (string (fourth data)))))
       
      (:mine-true
       (mod-chunk-fct 'cond1-loc (list 'visible t))
       (setf (chunk-real-visual-value 'cond1-loc) "+mine")
       (mod-chunk-fct 'cond1 (list 'value "+mine"))
       )
      (:mine-false
       (mod-chunk-fct 'cond1-loc (list 'visible t))
       (setf (chunk-real-visual-value 'cond1-loc) "-mine")
       (mod-chunk-fct 'cond1 (list 'value "-mine"))
       (setf (mine-visible world) nil)
       )
      (:fortress-true
       (mod-chunk-fct 'cond2-loc (list 'visible t))
       (setf (chunk-real-visual-value 'cond2-loc) "+fort")
       (mod-chunk-fct 'cond2 (list 'value "+fort"))
       (setf (fort-exists world) t)
       )
      (:fortress-false
       (mod-chunk-fct 'cond2-loc (list 'visible t))
       (setf (chunk-real-visual-value 'cond2-loc) "-fort")
       (mod-chunk-fct 'cond2 (list 'value "-fort"))
       (setf (fort-exists world) nil))
      ))
  
  ;;; a fortress-hit stays "visible" for 3 ticks ~100ms
  ;;; may also want to use an audio cue for that too since
  ;;; the game does make a sound, but for now just keeping
  ;;; it visual
  
  (when (numberp (f-hit-counter world))
    (incf (f-hit-counter world))
    
    (if (or (not (fast-chunk-slot-value-fct 'fortress-loc 'visible)) (= (f-hit-counter world) 4))
        (setf (f-hit-counter world) nil)
      (mod-chunk-fct 'fortress-hit-loc (list 'visible t))
      ))
  
    (when (and (fortress-there world) (not (fast-chunk-slot-value-fct 'fortress-loc 'visible)))
      (mod-chunk-fct 'fortress-explosion-loc (list 'visible t)))
  
    (when (and (mine-visible world)
               (null (fast-chunk-slot-value-fct 'mine-loc 'visible))
               (not (zerop (mine-visible world))))
      (mod-chunk-fct 'mine-loc (list 'visible t 'kind 'dead-mine))
      (decf (mine-visible world)))
    
    
  (setf (new-data world) nil)
  (when (last-proc-handled world)
    (setf (last-proc-handled world) nil)
    (proc-display))))



;;; functions to compute the simulation of the player shooting at the mine
;;; used by the imaginal module.

;;; Assume the model won't thrust.

;;; Compute the number of game ticks it would take to hit the mine 
;;; - if it doesn't turn
;;; - if it turns to the side the mine is on
;;;      compute the shortest to hit and the number of angle steps
;;;
;;; 
;;; Assume there's a 9 tick delay before any turning will occur and that
;;; the my-ship and mine chunks hold the necessary info.
;;; 
;;; If the it can shoot the mine without turning in less that 5 ticks
;;; it won't bother to compute the others and just go with that.
;;; 

(defvar *show-prediction* nil)


(defun predict-mine-shooting ()
  (let* ((vx (chunk-slot-value my-ship vx))
         (vy (chunk-slot-value my-ship vy))
         (mx (chunk-slot-value mine x))
         (my (chunk-slot-value mine y))
         (o (chunk-slot-value my-ship orientation))
         (sx (chunk-slot-value my-ship x))
         (sy (chunk-slot-value my-ship y))
         (side nil)
        ; (mdist (dist (list mx my) (list sx sy)))
        ; (hit-dist (abs (* (sin (deg->rad a)) dist)))
         (fa (chunk-slot-value my-ship angle))
         (fd (chunk-slot-value my-ship dist))
        ; (dfa (- o fa))
        ; (hit-f-dist (abs (* (sin (deg->rad dfa)) fd)))
         (no-turn-steps nil)
         (turn-steps nil)
         (turns nil)
         (d-hit-dist nil)
         (t-hit-dist nil)
         )
    
    ;;; compute the number of steps needed to drift
    
    (do* ((count 0 (1+ count))
          (csx sx (+ csx vx))
          (csy sy (- csy vy))
          (mh 0 (atan (- cmy csy) (- csx cmx)))
          (cmx mx (+ cmx (* (cos mh) 5)))
          (cmy my (- cmy (* (sin mh) 5)))
          (cfa fa (- o (rad->deg (atan (- csy 315) (- 355 csx)))))
          (cfd fd (dist (list csx csy) '(355 315)))
          
          (hit-list (multiple-value-list (can-shoot-mine cmx cmy csx csy o cfa cfd))
                    (multiple-value-list (can-shoot-mine cmx cmy csx csy o cfa cfd)))
          (hit (car hit-list) (car hit-list))
          (done (second hit-list) (second hit-list))
          
          )
         ((or hit done) (when hit (setf no-turn-steps count)
                          (setf d-hit-dist (dist (list cmx cmy) (list csx csy)))))
      
      
      )
    
    (if (and no-turn-steps (<= no-turn-steps 5))
        (set-chunk-slot-value mine-predict action drift)
      
      ;; compute the steps if the model turns
      (do* ((count 0 (1+ count))
            (csx sx (+ csx vx))
            (csy sy (- csy vy))
            (mh 0 (atan (- cmy csy) (- csx cmx)))
            (cmx mx (+ cmx (* (cos mh) 5)))
            (cmy my (- cmy (* (sin mh) 5)))
            
            (o o)
            (cfa fa)
            (cfd fd (dist (list csx csy) '(355 315)))
            (hit-list (multiple-value-list (can-shoot-mine cmx cmy csx csy o cfa cfd)))
            (hit (car hit-list))
            (done (second hit-list)))
           ((or hit done)
            (when hit 
              (setf turn-steps count)
              (setf t-hit-dist (dist (list cmx cmy) (list csx csy)))))
        
        (if (<= count 9)
            (multiple-value-bind (h d) (can-shoot-mine cmx cmy csx csy o cfa cfd)
              (setf ; hit h if it would shoot in less than 9 it's already covered
                    ; and otherwise I need to recompute the aim angle to check the hit
               done d))
          
          (do* ;; iterate from 1 to 30 turns (or max that could have been made)
               ;; check to see if any hit or ship has been hit
               ;; if so return and then have those values recorded...
               ((turn 0 (1+ turn))
                (co-left (+ o (* turn 6)) (+ o (* turn 6)))
                (co-right (- o (* turn 6)) (- o (* turn 6)))
                (tfa-left (- co-left (rad->deg (atan (- csy 315) (- 355 csx)))) 
                          (- co-left (rad->deg (atan (- csy 315) (- 355 csx)))))
                (tfa-right (- co-right (rad->deg (atan (- csy 315) (- 355 csx)))) 
                           (- co-right (rad->deg (atan (- csy 315) (- 355 csx)))))
                (t-hit-list-left (multiple-value-list (can-shoot-mine cmx cmy csx csy co-left tfa-left cfd))
                                 (multiple-value-list (can-shoot-mine cmx cmy csx csy co-left tfa-left cfd)))
                (t-hit-list-right (multiple-value-list (can-shoot-mine cmx cmy csx csy co-right tfa-right cfd))
                                  (multiple-value-list (can-shoot-mine cmx cmy csx csy co-right tfa-right cfd)))
                (t-hit (or (car t-hit-list-left) (car t-hit-list-right)) 
                       (or (car t-hit-list-left) (car t-hit-list-right)))
                (t-done (or (second t-hit-list-left) (second t-hit-list-right)) 
                        (or (second t-hit-list-left) (second t-hit-list-right))))
               ((or t-hit t-done (> turn (- count 9)) (> turn 30))
                (when (or t-hit t-done) 
                  (setf hit t-hit 
                    done t-done
                    side (if (car t-hit-list-left) 'left 'right))
                  (when t-hit
                    (setf turns turn))))
            
            ;(format t "~d ~d ~3s~%" count turn t-hit)
            ))
            
            
            
        ))
    
    
    (when *show-prediction* 
      (format t "No turn: ~s  turn: ~s  steps: ~s side: ~s" no-turn-steps turn-steps turns side))
          
    (cond ((null (or no-turn-steps turn-steps))
           ;; can't hit it in time...
           (set-chunk-slot-value mine-predict action panic)
           )
          ((and no-turn-steps turn-steps)
           ;; hit it with either -- give a preference to drifting
           ;; if it's close in time but if turning is a clear
           ;; winner go with it
           (if (<= no-turn-steps (+ turn-steps 3))
               (progn
                 (set-chunk-slot-value-fct 'mine-predict 'ticks (compute-ticks no-turn-steps))
                 (set-chunk-slot-value mine-predict action drift)
                 (set-chunk-slot-value-fct 'mine-predict 'dist (+ d-hit-dist 40))
                 (set-chunk-slot-value-fct 'mine-predict 'hit-dist (+ d-hit-dist -90))
                 
                 ) ;; give it about 5 extra ticks
             (progn
               (set-chunk-slot-value mine-predict action turn)
               (if (eq side 'right)
                   (set-chunk-slot-value mine-predict finger index)
                 (set-chunk-slot-value mine-predict finger ring))
               (set-chunk-slot-value-fct 'mine-predict 'ticks (compute-ticks turn-steps))
               (set-chunk-slot-value-fct 'mine-predict 'delay (* turns .033))
               (set-chunk-slot-value-fct 'mine-predict 'hit-dist (+ t-hit-dist -90))
               (set-chunk-slot-value-fct 'mine-predict 'dist (+ t-hit-dist 40 (* 5 turns))) ;; buffer of 5 ticks (of mine speed) plus the turn time
                                                                                                 ;; with a slack for the event that the ship and mine
                                                                                                 ;; are moving toward each other.
               )))
          (no-turn-steps
           (set-chunk-slot-value mine-predict action drift)
           (set-chunk-slot-value-fct 'mine-predict 'ticks (compute-ticks no-turn-steps))
           (set-chunk-slot-value-fct 'mine-predict 'hit-dist (+ d-hit-dist -90))
           (set-chunk-slot-value-fct 'mine-predict 'dist (+ d-hit-dist 40)))
          (turn-steps
           (set-chunk-slot-value mine-predict action turn)
           (set-chunk-slot-value-fct 'mine-predict 'ticks (compute-ticks turn-steps))
           (if (eq side 'right)
               (set-chunk-slot-value mine-predict finger index)
             (set-chunk-slot-value mine-predict finger ring))
           (set-chunk-slot-value-fct 'mine-predict 'delay (* turns .033))
           (set-chunk-slot-value-fct 'mine-predict 'hit-dist (+ t-hit-dist -90))
           (set-chunk-slot-value-fct 'mine-predict 'dist (+ t-hit-dist 40 (* 5 turns))))
          (t (model-warning "Prediction code failed...")))
    
    
    (schedule-set-buffer-chunk 'imaginal 'mine-predict .2 :module 'imaginal)
    
    (schedule-event-relative .2 'set-imaginal-free :module 'imaginal)
    ))

(defun compute-ticks (steps)
  (let ((time (* steps .033)))
    (1+ (ceiling (1- (log (1+ (/ time .11)) 1.1))))))

;; returns 2 values: ship-can-shoot-mine and mine-hits-ship

(defun can-shoot-mine (mx my sx sy o fa fd)
  
  (when *show-prediction*
        (format t "s: ~6,3f ~6,3f m: ~6,3f ~6,3f o: ~6,3f fa: ~6,3f"
          sx sy mx my o fa))
  
  (if (or (< sx 0) (< sy 0) (> sx 710) (> sy 626))
      (progn
        (when *show-prediction*
          (format t "nil t~%"))
        
        (values nil t))
    
    (let* ((ma (rad->deg (atan (- sy my) (- mx sx))))
           (a (- o ma))
           (adjust-a (if (> a 180)
                         (- a 360)
                       (if (< a -180)
                           (+ a 360)
                         a)))
            
           (mdist (dist (list mx my) (list sx sy)))
           (hit-dist (abs (* (sin (deg->rad a)) mdist)))
           (shoot (and (<= hit-dist 20) (< (abs adjust-a) 90)
                   (not (and ;(fort-exists (current-device)) 
                         (< fd mdist)
                         (<= (abs (* (sin (deg->rad fa)) fd)) 18)))))
           (done (< mdist 30)))
      
      (when *show-prediction*
        (format t "~s ~S~%" shoot done))
     
      (values shoot done))))


;;; compute turn and thrust to get back into orbit after
;;; handling a mine
;;;
;;; Using a speed based correction
;;;  - low speed (< 1)
;;;    turn out to the tangent angle
;;;    give a fast thrust
;;;  - medium speed (1 - 1.5)
;;;    turn to 1/2 tangent angle 
;;;    give a fast thrust
;;;  - high speed (1.5 - 1.9)
;;;    turn to 0
;;;  - give a default thrust
;;;  - very high speed > 1.9
;;;    turn to - tangent angle 
;;;  - give a default thrust
;;; 
;;; Assume that the model is already turning to
;;; near 0 if the |angle| is > 20.

(defun reorbit-from-dist ()
  
  (let* ((h (chunk-slot-value my-ship dist))
         (angle (chunk-slot-value my-ship angle))
         (r 90)
         (speed (chunk-slot-value my-ship vel))
         (theta (rad->deg (asin (/ r h))))
         ;(vdir (chunk-slot-value my-ship vdir))
         (thrust-len (if (< speed 1.5) 'fast 'default))
         (target-angle (cond ((< speed 1) theta)
                             ((< speed 1.5) (/ theta 2))
                             (t 0)
                             ))
         )
    
    
    (unless (<= -20 angle 20)
      (setf angle 0))
    
    (let ((delta-angle (- angle target-angle))
          (turn-time nil)
          (turn-finger nil))
      
      (when (> delta-angle 6)
        (setf turn-time (* .033 (round (abs delta-angle) 6)))
        (setf turn-finger (if (plusp delta-angle) 'index 'ring)))
        
      
      (set-chunk-slot-value-fct 'reorbit 'turn turn-time)
      (set-chunk-slot-value-fct 'reorbit 'finger turn-finger)

      (set-chunk-slot-value-fct 'reorbit 'thrust thrust-len)
    
      
      ;(format t "theta: ~s Target-angle: ~s  delta-angle: ~s~%" theta target-angle delta-angle)
      ;(pprint-chunks my-ship reorbit)
    
    
    (schedule-set-buffer-chunk 'imaginal 'reorbit .2 :module 'imaginal)
    
    (schedule-event-relative .2 'set-imaginal-free :module 'imaginal))))




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Function to handle the real work


(defun collect-sf-incoming-data (world)
  (unwind-protect 
      (loop
        (multiple-value-bind (state-line err)
            (ignore-errors (uni-socket-read-line (socket world)))
          (if (subtypep (type-of err) 'condition)
              (return)
            (unless (string= "" state-line)
              
              ;; assume the data is all readable items and just create a list of them to parse
              
              (let ((data (read-from-string (concatenate 'string "(" state-line ")"))))                  
                (when (collect-incoming world)
                  (push (list (get-internal-real-time) data) (incoming-record world)))
                
                
                ;;; parse the data and modify the location and object chunks directly...
                (case (car data)
                  (:start (setf (end-time world) (/ (second data) 1000.0)))
                  
                  (t (push data (new-data world)))))))))
    
    (setf (game-state world) 'disconnected)))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The module to provide the interface, parameters, etc

(defun create-sf-module (name)
  (declare (ignore name))
;  (add-post-event-hook 'sf-post-hook)
  (make-instance 'sf-world))




(defun pre-reset-sf-module (world)
  
  ;; Kill any previous connection
  
  (when (socket world)
    (close (socket world))
    (setf (socket world) nil))
  
  ;; kill any existing socket handler
  
  (when (state-process world)
    (uni-process-kill (state-process world))
    (setf (state-process world) nil))
  
  ;; clear out the slots as needed
 
  (setf (game-state world) 'disconnected)
  (clrhash (loc->obj-table world))
  
  (setf (last-proc-handled world) t)
  (setf (proc-now world) nil)
  (setf (new-data world) nil)
  (setf (incoming-record world) nil)
  (setf (outgoing-record world) nil)
  (setf (end-time world) 0)
  
  (setf (f-hit-counter world) nil)
  (setf (hit-hex world) nil)
  (setf (last-v world) nil)
  (setf (mine-visible world) nil)
  
  ;; define the chunk-types for the requests
  ;; used 
  
  (chunk-type connect)
  )

  
(defun post-reset-sf-module (world)
  
  ;; create special visual-location chunk-type for the items
  ;; with any extra slots needed
  ;; multiple subtypes may be useful
  
  (chunk-type (sf-visual-location (:include visual-location))
	      visible)
  
  
  ;; same for visual-objects
  
  (chunk-type (sf-object (:include visual-object)))
  
  
  (chunk-type (ship (:include sf-object))
	      x y angle vel dist vdir hex-dist orientation last-dist hex-hit mine-side mine-dist shoot-at-mine hit-fortress-first vx vy speed-change)
  
  (chunk-type (mine (:include sf-object)) x y)
  
  (chunk-type (fortress (:include sf-object)) angle)
  (chunk-type (bonus (:include sf-object)))
  (chunk-type (trial-end (:include sf-object)))
  (chunk-type mine-heading action finger delay ticks dist hit-dist)
  (chunk-type reorbit turn thrust finger)
  
  
  (define-chunks (ship isa chunk)
      (own-ship-loc isa sf-visual-location kind ship)
    (explosion-loc isa sf-visual-location kind explosion)
    (fortress-loc isa sf-visual-location kind fortress screen-x 355 screen-y 315)
    (fortress-explosion-loc isa sf-visual-location kind fortress-explosion screen-x 355 screen-y 315)
    (bonus-loc isa sf-visual-location kind bonus screen-x 355 screen-y 420)
    (vlner-loc isa sf-visual-location screen-x  465 screen-y 700 kind text)
    (fortress-hit-loc isa sf-visual-location kind fortress-hit screen-x 355 screen-y 315)
    (trial-end-loc isa sf-visual-location kind trial-end screen-x 355 screen-y 315)
    (foe1-loc isa sf-visual-location screen-x 200 screen-y 500 kind text)
    (foe2-loc isa sf-visual-location screen-x 300 screen-y 500 kind text)
    (foe3-loc isa sf-visual-location screen-x 400 screen-y 500 kind text)
    
    (cond1-loc isa sf-visual-location screen-x 200 screen-y 500 kind text)
    (cond2-loc isa sf-visual-location screen-x 400 screen-y 500 kind text)
    (iff-loc isa sf-visual-location screen-x 600 screen-y 700 kind text)
    (mine-loc isa sf-visual-location kind mine)
    
    (mine-predict isa mine-heading)
    
    (reorbit isa reorbit)
    
    (new-trial-end isa chunk)
    
    (my-ship isa ship)
    (explosion isa sf-object value explosion)
    (fortress isa sf-object)
    (fortress-explosion isa sf-object)                        
    (bonus isa bonus)
    (vlner isa text value 0)
    (fortress-hit isa sf-object)
    (trial-end isa trial-end)
    
    (foe1 isa text)
    (foe2 isa text)
    (foe3 isa text)
    (cond1 isa text)
    (cond2 isa text)
    
    (iff isa text)
    (mine isa mine)
    (debris isa chunk)
    )
  
  ;; create all the visual-location and visual-object
  ;; chunks up front and set up the mappings - assuming
  ;; the locs and objs are created in corresponding order
  
  (setf (dynamics world) 
    (list 'own-ship-loc 'explosion-loc 'fortress-loc 'fortress-explosion-loc 'bonus-loc 'fortress-hit-loc 'trial-end-loc 'vlner-loc
          'foe1-loc 'foe2-loc 'foe3-loc 'cond1-loc 'cond2-loc 'iff-loc 'mine-loc))
    
  
  (setf (all-locs world) (dynamics world))
  
    ;(append (dynamics world) (list 'vlner-loc)))
  
  (mapc (lambda (loc obj) (setf (gethash loc (loc->obj-table world)) obj))
    (all-locs world)
    (list 'my-ship 'explosion 'fortress 'fortress-explosion 'bonus 'fortress-hit 'trial-end 'vlner 'foe1 'foe2 'foe3 'cond1 'cond2 'iff 'mine))
          
          
  ;; automatically make this the current device
  
  (install-device world))


(defun delete-sf-module (world)
  ;; Kill any open connections
  
  (when (socket world)
    (close (socket world))
    (setf (socket world) nil))
  
  (when (state-process world)
    (uni-process-kill (state-process world))
    (setf (state-process world) nil)))


;; respond to the game buffer queries

(defun sf-query (world buffer-name slot value)
  (case buffer-name
    
    (game (case slot
            (state
             (case value
               (busy nil)
               (free t)
               (error nil)
               (t (print-warning "Unknown state query ~S to game buffer" value)
                  nil)))
            (game-state
                (eq (game-state world) value))
            (t (print-warning "Unknown query ~S ~S to the game buffer" slot value))))))



;; handle the buffer requests (game buffer connect being the only option at this poing)

(defun sf-request (world buffer-name spec)
  (case buffer-name
    (game
     (case (chunk-spec-chunk-type spec)
       
       (t (print-warning 
           "Invalid command request ~S to the game buffer" (chunk-spec-chunk-type spec)))))))



;; custom function for printing the buffer-status of the game buffer

(defun print-sf-game-state ()
  (command-output "  game-state connecting  : ~S"
                  (query-buffer 'game
                                '((game-state . connecting))))
  (command-output "  game-state connected   : ~S"
                  (query-buffer 'game
                                '((game-state . connected))))
  (command-output "  game-state disconnected: ~S"
                  (query-buffer 'game
                                '((game-state . disconnected)))))


;; function to handle the setting and getting of the module's parameters

(defun sf-params (world param)
  (cond ((consp param)
         
          (case (car param)
            (:sf-host (setf (sf-host world) (cdr param)))
            (:sf-port (setf (sf-port world) (cdr param)))
            (:log-sf-data-in (setf (collect-incoming world) (cdr param)))
            (:log-sf-data-out (setf (record-outgoing world) (cdr param)))))
        (t 
         (case param
           (:sf-host (sf-host world))
           (:sf-port (sf-port world))
           (:log-sf-data-in (collect-incoming world))
           (:log-sf-data-out (record-outgoing world))))))


;; define the module itself

(define-module-fct :sf '((game nil nil (game-state) print-sf-game-state)) 
  (list
   (define-parameter :sf-host :valid-test #'stringp :default-value *host-address*
     :warning "a string" :documentation "IP address or host name of SF server")
   (define-parameter :sf-port :valid-test 'integerp  :default-value *port-for-server*
     :warning "an integer" :documentation "Gamebots port of the server for connecting")
   (define-parameter :log-sf-data-in :valid-test #'tornil :default-value nil
     :warning "T or nil" :documentation "Record all incoming SF data")
   (define-parameter :log-sf-data-out :valid-test #'tornil :default-value nil
     :warning "T or nil" :documentation "Record all commands sent to SF"))
  
  :version "0.1"
  :documentation "Module to allow ACT-R to connect to SpaceFortress with our custom socket interface"
  :creation 'create-sf-module
  :reset (list 'pre-reset-sf-module 'post-reset-sf-module)
  :delete 'delete-sf-module
  :request 'sf-request
  :query 'sf-query
  :params 'sf-params)


(defun connect-to-game ()
  (let ((world (get-module :sf)))
    (if (not (eq (game-state world) 'disconnected))
        
        (print-warning "Model is already connected cannot connect.")
      
      (progn
        (setf (socket world)
          (uni-make-socket (sf-host world) (sf-port world)))
        
        (setf (end-time world) nil)
        
        (if (socket world)
            (progn
              (setf (game-state world) 'connecting)
              (setf (state-process world)
                (uni-run-process 
                 "SF state harvester"
                 (lambda ()
                   (collect-sf-incoming-data world))))
              
              ;; Make sure we get the first update
              ;; before progressing.
              
              (while (null (end-time world)) (uni-process-system-events))
              
              (setf (game-state world) 'connected))
              
          (print-warning "Could not open socket connection."))))))

;; The function to call to run things
;; Only runs in real-time since it needs to stay synced to the 
;; game's clock -- to get as fast as possible just provide a
;; large speed value.

(defun play-sf (time &key (speed 1) (cont nil))
  
  (let ((module (get-module :sf)))
    
    (setf (start-time module) nil)
    (setf (sim-speed module) (/ 1.0 speed))
  
    
    (if cont
        (progn
          (setf (last-stop-time module) (mp-time))
          (run-full-time time :real-time t))
      (progn
        (reset)
         
        (connect-to-game)
        
        (setf (last-stop-time module) 0.0)
       
        (mp-real-time-management ; not using a :max-time-delta since the slack does a proc-display
                                 :units-per-second 1
                                 :slack-function 'sf-slack
                                 :time-function 'sf-time
                                 :allow-dynamics t)
        
        ;; update with the initial info right away
        
       ; (schedule-event 0.0 'update-world :destination :sf :module :sf :maintenance t :priority :max)
        
        (run-full-time time :real-time t)))))


(defun sf-time ()
  (let ((module (get-module :sf)))
    (if (null (start-time module))
        (progn
          (setf (start-time module) (get-internal-real-time))
          (+ 0.0 (last-stop-time module)))
      (let* ((current (+ (last-stop-time module) (/ (- (get-internal-real-time) (start-time module)) internal-time-units-per-second (sim-speed module)))))
        (cond 
         ((> current (end-time module))
          (setf (last-time module) :ready)
          (end-time module))
         (t
          current))))))
    
    
(defun sf-slack (delta)
  (let* ((module (get-module :sf))
         (end (end-time module)))
    ;(format t "~15s delta:~9,3f end:~9,3f mp-time:~9,3f~%" (last-time module) delta end (mp-time))
    (when (eq :ready (last-time module))
      (socket-write-line module "ready")
      (setf (last-time module) :not-ready)
      
      (while (= end (end-time module))
        (uni-process-system-events))
      
      #|(if ;(> (+ end delta) (mp-time))
          ; That's not useful because it's never going to be
          ; true -- at best they'll be equal.
          ; What I care about is whether or not the
          ; model has an event before the next time I
          ; need a sf update.  If there's no model event
          ; before then I can skip to the next sf time
          ; without updating the world.
          (<= (end-time module) (mp-time))
          
          (progn
            ;; clear out the current update data
            ;; and skip to the next update
            (setf (last-time module) :ready)
            (setf (new-data module) nil)
            (sf-slack (- (+ end delta) (end-time module))))
        (update-world module))
      
      |#
      
      (schedule-event end 'update-world :destination :sf :module :sf :maintenance nil)
      
      
      )))




      

(unless (boundp '*loaded-modules*)
  (defparameter *loaded-modules* t)
  (compile-and-load "ACT-R6:speculative-modules;key-hold;key-press-module.lisp")
  (compile-and-load "ACT-R6:speculative-modules;key-hold;delayed-punch.lisp")
  (compile-and-load "ACT-R6:speculative-modules;key-hold;hold-until.lisp")
  (compile-and-load "ACT-R6:speculative-modules;visual-search;visual-search-buffer.lisp")
  (compile-and-load "ACT-R6:speculative-modules;attend-and-track;attend-and-track.lisp")
  (compile-and-load "ACT-R6:speculative-modules;two-handed;two-exec-paths.lisp"))
  



#|
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
|#
